@@ -4,28 +4,28 @@ Make beautiful, interactive maps with Python and Leaflet.js
 """
 
 import time
-import warnings
 import webbrowser
+from typing import Any, List, Optional, Sequence, Union
 
-from branca.element import Element, Figure, MacroElement
-from jinja2 import Environment, PackageLoader, Template
+from branca.element import Element, Figure
+from jinja2 import Template
 
 from folium.elements import JSCSSMixin
-from folium.map import FitBounds
+from folium.map import Evented, FitBounds, Layer
 from folium.raster_layers import TileLayer
 from folium.utilities import (
+    TypeBounds,
+    TypeJsonValue,
     _parse_size,
+    parse_font_size,
     parse_options,
     temp_html_filepath,
     validate_location,
 )
 
-ENV = Environment(loader=PackageLoader("folium", "templates"))
-
-
 _default_js = [
     ("leaflet", "https://cdn.jsdelivr.net/npm/leaflet@1.9.3/dist/leaflet.js"),
-    ("jquery", "https://code.jquery.com/jquery-1.12.4.min.js"),
+    ("jquery", "https://code.jquery.com/jquery-3.7.1.min.js"),
     (
         "bootstrap",
         "https://cdn.jsdelivr.net/npm/bootstrap@5.2.2/dist/js/bootstrap.bundle.min.js",
@@ -45,7 +45,7 @@ _default_css = [
     # glyphicons came from Bootstrap 3 and are used for Awesome Markers
     (
         "glyphicons_css",
-        "https://netdna.bootstrapcdn.com/bootstrap/3.0.0/css/bootstrap.min.css",
+        "https://netdna.bootstrapcdn.com/bootstrap/3.0.0/css/bootstrap-glyphicons.css",
     ),
     (
         "awesome_markers_font_css",
@@ -63,7 +63,6 @@ _default_css = [
 
 
 class GlobalSwitches(Element):
-
     _template = Template(
         """
         <script>
@@ -80,29 +79,24 @@ class GlobalSwitches(Element):
         self.disable_3d = disable_3d
 
 
-class Map(JSCSSMixin, MacroElement):
+class Map(JSCSSMixin, Evented):
     """Create a Map with Folium and Leaflet.js
 
     Generate a base map of given width and height with either default
-    tilesets or a custom tileset URL. The following tilesets are built-in
-    to Folium. Pass any of the following to the "tiles" keyword:
+    tilesets or a custom tileset URL. Folium has built-in all tilesets
+    available in the ``xyzservices`` package. For example, you can pass
+    any of the following to the "tiles" keyword:
 
         - "OpenStreetMap"
-        - "Mapbox Bright" (Limited levels of zoom for free tiles)
-        - "Mapbox Control Room" (Limited levels of zoom for free tiles)
-        - "Stamen" (Terrain, Toner, and Watercolor)
-        - "Cloudmade" (Must pass API key)
-        - "Mapbox" (Must pass API key)
-        - "CartoDB" (positron and dark_matter)
+        - "CartoDB Positron"
+        - "CartoDB Voyager"
 
-    You can pass a custom tileset to Folium by passing a
+    Explore more provider names available in ``xyzservices`` here:
+    https://leaflet-extras.github.io/leaflet-providers/preview/.
+
+    You can also pass a custom tileset by passing a
     :class:`xyzservices.TileProvider` or a Leaflet-style
-    URL to the tiles parameter: ``http://{s}.yourtiles.com/{z}/{x}/{y}.png``.
-
-    You can find a list of free tile providers here:
-    ``http://leaflet-extras.github.io/leaflet-providers/preview/``.
-    Be sure to check their terms and conditions and to provide attribution
-    with the `attr` keyword.
+    URL to the tiles parameter: ``https://{s}.yourtiles.com/{z}/{x}/{y}.png``.
 
     Parameters
     ----------
@@ -118,10 +112,12 @@ class Map(JSCSSMixin, MacroElement):
         pass a custom URL, pass a TileLayer object,
         or pass `None` to create a map without tiles.
         For more advanced tile layer options, use the `TileLayer` class.
-    min_zoom: int, default 0
+    min_zoom: int, optional, default 0
         Minimum allowed zoom level for the tile layer that is created.
-    max_zoom: int, default 18
+        Filled by xyzservices by default.
+    max_zoom: int, optional, default 18
         Maximum allowed zoom level for the tile layer that is created.
+        Filled by xyzservices by default.
     zoom_start: int, default 10
         Initial zoom level for the map.
     attr: string, default None
@@ -154,8 +150,12 @@ class Map(JSCSSMixin, MacroElement):
         Forces Leaflet to not use hardware-accelerated CSS 3D
         transforms for positioning (which may cause glitches in some
         rare environments) even if they're supported.
-    zoom_control : bool, default True
-        Display zoom controls on the map.
+    zoom_control : bool or position string, default True
+        Display zoom controls on the map. The default `True` places it in the top left corner.
+        Other options are 'topleft', 'topright', 'bottomleft' or 'bottomright'.
+    font_size : int or float or string (default: '1rem')
+        The font size to use for Leaflet, can either be a number or a
+        string ending in 'rem', 'em', or 'px'.
     **kwargs
         Additional keyword arguments are passed to Leaflets Map class:
         https://leafletjs.com/reference.html#map
@@ -190,7 +190,7 @@ class Map(JSCSSMixin, MacroElement):
                     left: {{this.left[0]}}{{this.left[1]}};
                     top: {{this.top[0]}}{{this.top[1]}};
                 }
-                .leaflet-container { font-size: 1rem; }
+                .leaflet-container { font-size: {{this.font_size}}; }
             </style>
         {% endmacro %}
 
@@ -214,6 +214,10 @@ class Map(JSCSSMixin, MacroElement):
             L.control.scale().addTo({{ this.get_name() }});
             {%- endif %}
 
+            {%- if this.zoom_control_position %}
+            L.control.zoom( { position: {{ this.zoom_control|tojson }} } ).addTo({{ this.get_name() }});
+            {%- endif %}
+
             {% if this.objects_to_stay_in_front %}
             function objects_in_front() {
                 {%- for obj in this.objects_to_stay_in_front %}
@@ -234,41 +238,41 @@ class Map(JSCSSMixin, MacroElement):
 
     def __init__(
         self,
-        location=None,
-        width="100%",
-        height="100%",
-        left="0%",
-        top="0%",
-        position="relative",
-        tiles="OpenStreetMap",
-        attr=None,
-        min_zoom=0,
-        max_zoom=18,
-        zoom_start=10,
-        min_lat=-90,
-        max_lat=90,
-        min_lon=-180,
-        max_lon=180,
-        max_bounds=False,
-        crs="EPSG3857",
-        control_scale=False,
-        prefer_canvas=False,
-        no_touch=False,
-        disable_3d=False,
-        png_enabled=False,
-        zoom_control=True,
-        **kwargs,
+        location: Optional[Sequence[float]] = None,
+        width: Union[str, float] = "100%",
+        height: Union[str, float] = "100%",
+        left: Union[str, float] = "0%",
+        top: Union[str, float] = "0%",
+        position: str = "relative",
+        tiles: Union[str, TileLayer, None] = "OpenStreetMap",
+        attr: Optional[str] = None,
+        min_zoom: Optional[int] = None,
+        max_zoom: Optional[int] = None,
+        zoom_start: int = 10,
+        min_lat: float = -90,
+        max_lat: float = 90,
+        min_lon: float = -180,
+        max_lon: float = 180,
+        max_bounds: bool = False,
+        crs: str = "EPSG3857",
+        control_scale: bool = False,
+        prefer_canvas: bool = False,
+        no_touch: bool = False,
+        disable_3d: bool = False,
+        png_enabled: bool = False,
+        zoom_control: Union[bool, str] = True,
+        font_size: str = "1rem",
+        **kwargs: TypeJsonValue,
     ):
         super().__init__()
         self._name = "Map"
-        self._env = ENV
-        # Undocumented for now b/c this will be subject to a re-factor soon.
-        self._png_image = None
+
+        self._png_image: Optional[bytes] = None
         self.png_enabled = png_enabled
 
         if location is None:
             # If location is not passed we center and zoom out.
-            self.location = [0, 0]
+            self.location = [0.0, 0.0]
             zoom_start = 1
         else:
             self.location = validate_location(location)
@@ -281,6 +285,7 @@ class Map(JSCSSMixin, MacroElement):
         self.left = _parse_size(left)
         self.top = _parse_size(top)
         self.position = position
+        self.font_size = parse_font_size(font_size)
 
         max_bounds_array = (
             [[min_lat, min_lon], [max_lat, max_lon]] if max_bounds else None
@@ -289,17 +294,28 @@ class Map(JSCSSMixin, MacroElement):
         self.crs = crs
         self.control_scale = control_scale
 
+        # Zoom control position specified ?
+        if isinstance(zoom_control, str):
+            self.zoom_control_position = True
+            if zoom_control not in {"topleft", "topright", "bottomleft", "bottomright"}:
+                raise ValueError(
+                    "Incorrect value for `zoom_control`, choose from 'topleft', 'topright', 'bottomleft' or 'bottomright'."
+                )
+            self.zoom_control = zoom_control
+        else:
+            self.zoom_control_position = False
+
         self.options = parse_options(
             max_bounds=max_bounds_array,
             zoom=zoom_start,
-            zoom_control=zoom_control,
+            zoom_control=False if self.zoom_control_position else zoom_control,
             prefer_canvas=prefer_canvas,
             **kwargs,
         )
 
         self.global_switches = GlobalSwitches(no_touch, disable_3d)
 
-        self.objects_to_stay_in_front = []
+        self.objects_to_stay_in_front: List[Layer] = []
 
         if isinstance(tiles, TileLayer):
             self.add_child(tiles)
@@ -309,17 +325,18 @@ class Map(JSCSSMixin, MacroElement):
             )
             self.add_child(tile_layer, name=tile_layer.tile_name)
 
-    def _repr_html_(self, **kwargs):
+    def _repr_html_(self, **kwargs) -> str:
         """Displays the HTML Map in a Jupyter notebook."""
         if self._parent is None:
             self.add_to(Figure())
+            self._parent: Figure
             out = self._parent._repr_html_(**kwargs)
             self._parent = None
         else:
             out = self._parent._repr_html_(**kwargs)
         return out
 
-    def _to_png(self, delay=3, driver=None):
+    def _to_png(self, delay: int = 3, driver: Any = None) -> bytes:
         """Export the HTML to byte representation of a PNG image.
 
         Uses selenium to render the HTML and record a PNG. You may need to
@@ -353,7 +370,7 @@ class Map(JSCSSMixin, MacroElement):
             self._png_image = png
         return self._png_image
 
-    def _repr_png_(self):
+    def _repr_png_(self) -> Optional[bytes]:
         """Displays the PNG Map in a Jupyter notebook."""
         # The notebook calls all _repr_*_ by default.
         # We don't want that here b/c this one is quite slow.
@@ -361,7 +378,7 @@ class Map(JSCSSMixin, MacroElement):
             return None
         return self._to_png()
 
-    def render(self, **kwargs):
+    def render(self, **kwargs) -> None:
         """Renders the HTML representation of the element."""
         figure = self.get_root()
         assert isinstance(
@@ -400,10 +417,10 @@ class Map(JSCSSMixin, MacroElement):
 
         super().render(**kwargs)
 
-    def show_in_browser(self):
+    def show_in_browser(self) -> None:
         """Display the Map in the default web browser."""
         with temp_html_filepath(self.get_root().render()) as fname:
-            webbrowser.open(fname)
+            webbrowser.open("file://" + fname)
             print(
                 "Your map should have been opened in your browser automatically."
                 "\nPress ctrl+c to return."
@@ -417,12 +434,12 @@ class Map(JSCSSMixin, MacroElement):
 
     def fit_bounds(
         self,
-        bounds,
-        padding_top_left=None,
-        padding_bottom_right=None,
-        padding=None,
-        max_zoom=None,
-    ):
+        bounds: TypeBounds,
+        padding_top_left: Optional[Sequence[float]] = None,
+        padding_bottom_right: Optional[Sequence[float]] = None,
+        padding: Optional[Sequence[float]] = None,
+        max_zoom: Optional[int] = None,
+    ) -> None:
         """Fit the map to contain a bounding box with the
         maximum zoom level possible.
 
@@ -457,22 +474,7 @@ class Map(JSCSSMixin, MacroElement):
             )
         )
 
-    def choropleth(self, *args, **kwargs):
-        """Call the Choropleth class with the same arguments.
-
-        This method may be deleted after a year from now (Nov 2018).
-        """
-        warnings.warn(
-            "The choropleth  method has been deprecated. Instead use the new "
-            "Choropleth class, which has the same arguments. See the example "
-            "notebook 'GeoJSON_and_choropleth' for how to do this.",
-            FutureWarning,
-        )
-        from folium.features import Choropleth
-
-        self.add_child(Choropleth(*args, **kwargs))
-
-    def keep_in_front(self, *args):
+    def keep_in_front(self, *args: Layer) -> None:
         """Pass one or multiple layers that must stay in front.
 
         The ordering matters, the last one is put on top.
