@@ -2,13 +2,19 @@
 Classes for drawing maps.
 
 """
+
 import warnings
 from collections import OrderedDict
+from typing import Dict, List, Optional, Sequence, Tuple, Type, Union
 
 from branca.element import Element, Figure, Html, MacroElement
 from jinja2 import Template
 
+from folium.elements import ElementAddToElement, EventHandler
 from folium.utilities import (
+    JsCode,
+    TypeBounds,
+    TypeJsonValue,
     camelize,
     escape_backticks,
     parse_options,
@@ -16,7 +22,23 @@ from folium.utilities import (
 )
 
 
-class Layer(MacroElement):
+class Evented(MacroElement):
+    """The base class for Layer and Map
+
+    Adds the `on` method for event handling capabilities.
+
+    See https://leafletjs.com/reference.html#evented for
+    more in depth documentation. Please note that we have
+    only added the `on(<Object> eventMap)` variant of this
+    method using python keyword arguments.
+    """
+
+    def on(self, **event_map: JsCode):
+        for event_type, handler in event_map.items():
+            self.add_child(EventHandler(event_type, handler))
+
+
+class Layer(Evented):
     """An abstract class for everything that is a Layer on the map.
     It will be used to define whether an object will be included in
     LayerControls.
@@ -30,16 +52,33 @@ class Layer(MacroElement):
     control : bool, default True
         Whether the Layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
     """
 
-    def __init__(self, name=None, overlay=False, control=True, show=True):
+    def __init__(
+        self,
+        name: Optional[str] = None,
+        overlay: bool = False,
+        control: bool = True,
+        show: bool = True,
+    ):
         super().__init__()
         self.layer_name = name if name is not None else self.get_name()
         self.overlay = overlay
         self.control = control
         self.show = show
 
+    def render(self, **kwargs):
+        if self.show:
+            self.add_child(
+                ElementAddToElement(
+                    element_name=self.get_name(),
+                    element_parent_name=self._parent.get_name(),
+                ),
+                name=self.get_name() + "_add",
+            )
+        super().render(**kwargs)
+
 
 class FeatureGroup(Layer):
     """
@@ -59,7 +98,7 @@ class FeatureGroup(Layer):
     control: bool, default True
         Whether the layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
     **kwargs
         Additional (possibly inherited) options. See
         https://leafletjs.com/reference.html#featuregroup
@@ -71,12 +110,19 @@ class FeatureGroup(Layer):
         {% macro script(this, kwargs) %}
             var {{ this.get_name() }} = L.featureGroup(
                 {{ this.options|tojson }}
-            ).addTo({{ this._parent.get_name() }});
+            );
         {% endmacro %}
         """
     )
 
-    def __init__(self, name=None, overlay=True, control=True, show=True, **kwargs):
+    def __init__(
+        self,
+        name: Optional[str] = None,
+        overlay: bool = True,
+        control: bool = True,
+        show: bool = True,
+        **kwargs: TypeJsonValue,
+    ):
         super().__init__(name=name, overlay=overlay, control=control, show=show)
         self._name = "FeatureGroup"
         self.tile_name = name if name is not None else self.get_name()
@@ -107,6 +153,9 @@ class LayerControl(MacroElement):
     autoZIndex : bool, default True
           If true the control assigns zIndexes in increasing order to all of
           its layers so that the order is preserved when switching them on/off.
+    draggable: bool, default False
+          By default the layer control has a fixed position. Set this argument
+          to True to allow dragging the control around.
     **kwargs
         Additional (possibly inherited) options. See
         https://leafletjs.com/reference.html#control-layers
@@ -116,7 +165,7 @@ class LayerControl(MacroElement):
     _template = Template(
         """
         {% macro script(this,kwargs) %}
-            var {{ this.get_name() }} = {
+            var {{ this.get_name() }}_layers = {
                 base_layers : {
                     {%- for key, val in this.base_layers.items() %}
                     {{ key|tojson }} : {{val}},
@@ -128,35 +177,42 @@ class LayerControl(MacroElement):
                     {%- endfor %}
                 },
             };
-            L.control.layers(
-                {{ this.get_name() }}.base_layers,
-                {{ this.get_name() }}.overlays,
+            let {{ this.get_name() }} = L.control.layers(
+                {{ this.get_name() }}_layers.base_layers,
+                {{ this.get_name() }}_layers.overlays,
                 {{ this.options|tojson }}
             ).addTo({{this._parent.get_name()}});
 
-            {%- for val in this.layers_untoggle.values() %}
-            {{ val }}.remove();
-            {%- endfor %}
+            {%- if this.draggable %}
+            new L.Draggable({{ this.get_name() }}.getContainer()).enable();
+            {%- endif %}
+
         {% endmacro %}
         """
     )
 
-    def __init__(self, position="topright", collapsed=True, autoZIndex=True, **kwargs):
+    def __init__(
+        self,
+        position: str = "topright",
+        collapsed: bool = True,
+        autoZIndex: bool = True,
+        draggable: bool = False,
+        **kwargs: TypeJsonValue,
+    ):
         super().__init__()
         self._name = "LayerControl"
         self.options = parse_options(
             position=position, collapsed=collapsed, autoZIndex=autoZIndex, **kwargs
         )
+        self.draggable = draggable
+        self.base_layers: OrderedDict[str, str] = OrderedDict()
+        self.overlays: OrderedDict[str, str] = OrderedDict()
+
+    def reset(self) -> None:
         self.base_layers = OrderedDict()
         self.overlays = OrderedDict()
-        self.layers_untoggle = OrderedDict()
 
-    def reset(self):
-        self.base_layers = OrderedDict()
-        self.overlays = OrderedDict()
-        self.layers_untoggle = OrderedDict()
-
-    def render(self, **kwargs):
+    def render(self, **kwargs) -> None:
         """Renders the HTML representation of the element."""
         self.reset()
         for item in self._parent._children.values():
@@ -165,12 +221,8 @@ class LayerControl(MacroElement):
             key = item.layer_name
             if not item.overlay:
                 self.base_layers[key] = item.get_name()
-                if len(self.base_layers) > 1:
-                    self.layers_untoggle[key] = item.get_name()
             else:
                 self.overlays[key] = item.get_name()
-                if not item.show:
-                    self.layers_untoggle[key] = item.get_name()
         super().render()
 
 
@@ -241,20 +293,18 @@ class Icon(MacroElement):
 
     def __init__(
         self,
-        color="blue",
-        icon_color="white",
-        icon="info-sign",
-        angle=0,
-        prefix="glyphicon",
-        **kwargs,
+        color: str = "blue",
+        icon_color: str = "white",
+        icon: str = "info-sign",
+        angle: int = 0,
+        prefix: str = "glyphicon",
+        **kwargs: TypeJsonValue,
     ):
         super().__init__()
         self._name = "Icon"
         if color not in self.color_options:
             warnings.warn(
-                "color argument of Icon should be one of: {}.".format(
-                    self.color_options
-                ),
+                f"color argument of Icon should be one of: {self.color_options}.",
                 stacklevel=2,
             )
         self.options = parse_options(
@@ -314,12 +364,12 @@ class Marker(MacroElement):
 
     def __init__(
         self,
-        location=None,
-        popup=None,
-        tooltip=None,
-        icon=None,
-        draggable=False,
-        **kwargs,
+        location: Optional[Sequence[float]] = None,
+        popup: Union["Popup", str, None] = None,
+        tooltip: Union["Tooltip", str, None] = None,
+        icon: Optional[Icon] = None,
+        draggable: bool = False,
+        **kwargs: TypeJsonValue,
     ):
         super().__init__()
         self._name = "Marker"
@@ -337,19 +387,18 @@ class Marker(MacroElement):
                 tooltip if isinstance(tooltip, Tooltip) else Tooltip(str(tooltip))
             )
 
-    def _get_self_bounds(self):
+    def _get_self_bounds(self) -> List[List[float]]:
         """Computes the bounds of the object itself.
 
         Because a marker has only single coordinates, we repeat them.
         """
+        assert self.location is not None
         return [self.location, self.location]
 
-    def render(self):
+    def render(self) -> None:
         if self.location is None:
             raise ValueError(
-                "{} location must be assigned when added directly to map.".format(
-                    self._name
-                )
+                f"{self._name} location must be assigned when added directly to map."
             )
         super().render()
 
@@ -399,13 +448,13 @@ class Popup(Element):
 
     def __init__(
         self,
-        html=None,
-        parse_html=False,
-        max_width="100%",
-        show=False,
-        sticky=False,
-        lazy=False,
-        **kwargs,
+        html: Union[str, Element, None] = None,
+        parse_html: bool = False,
+        max_width: Union[str, int] = "100%",
+        show: bool = False,
+        sticky: bool = False,
+        lazy: bool = False,
+        **kwargs: TypeJsonValue,
     ):
         super().__init__()
         self._name = "Popup"
@@ -434,7 +483,7 @@ class Popup(Element):
             **kwargs,
         )
 
-    def render(self, **kwargs):
+    def render(self, **kwargs) -> None:
         """Renders the HTML representation of the element."""
         for name, child in self._children.items():
             child.render(**kwargs)
@@ -482,7 +531,7 @@ class Tooltip(MacroElement):
         {% endmacro %}
         """
     )
-    valid_options = {
+    valid_options: Dict[str, Tuple[Type, ...]] = {
         "pane": (str,),
         "offset": (tuple,),
         "direction": (str,),
@@ -494,7 +543,13 @@ class Tooltip(MacroElement):
         "className": (str,),
     }
 
-    def __init__(self, text, style=None, sticky=True, **kwargs):
+    def __init__(
+        self,
+        text: str,
+        style: Optional[str] = None,
+        sticky: bool = True,
+        **kwargs: TypeJsonValue,
+    ):
         super().__init__()
         self._name = "Tooltip"
 
@@ -510,7 +565,10 @@ class Tooltip(MacroElement):
             # noqa outside of type checking.
             self.style = style
 
-    def parse_options(self, kwargs):
+    def parse_options(
+        self,
+        kwargs: Dict[str, TypeJsonValue],
+    ) -> Dict[str, TypeJsonValue]:
         """Validate the provided kwargs and return options as json string."""
         kwargs = {camelize(key): value for key, value in kwargs.items()}
         for key in kwargs.keys():
@@ -521,9 +579,7 @@ class Tooltip(MacroElement):
             )
             assert isinstance(
                 kwargs[key], self.valid_options[key]
-            ), "The option {} must be one of the following types: {}.".format(
-                key, self.valid_options[key]
-            )
+            ), f"The option {key} must be one of the following types: {self.valid_options[key]}."
         return kwargs
 
 
@@ -561,11 +617,11 @@ class FitBounds(MacroElement):
 
     def __init__(
         self,
-        bounds,
-        padding_top_left=None,
-        padding_bottom_right=None,
-        padding=None,
-        max_zoom=None,
+        bounds: TypeBounds,
+        padding_top_left: Optional[Sequence[float]] = None,
+        padding_bottom_right: Optional[Sequence[float]] = None,
+        padding: Optional[Sequence[float]] = None,
+        max_zoom: Optional[int] = None,
     ):
         super().__init__()
         self._name = "FitBounds"
@@ -578,6 +634,57 @@ class FitBounds(MacroElement):
         )
 
 
+class FitOverlays(MacroElement):
+    """Fit the bounds of the maps to the enabled overlays.
+
+    Parameters
+    ----------
+    padding: int, default 0
+        Amount of padding in pixels applied in the corners.
+    max_zoom: int, optional
+        The maximum possible zoom to use when fitting to the bounds.
+    fly: bool, default False
+        Use a smoother, longer animation.
+    fit_on_map_load: bool, default True
+        Apply the fit when initially loading the map.
+    """
+
+    _template = Template(
+        """
+        {% macro script(this, kwargs) %}
+        function customFlyToBounds() {
+            let bounds = L.latLngBounds([]);
+            {{ this._parent.get_name() }}.eachLayer(function(layer) {
+                if (typeof layer.getBounds === 'function') {
+                    bounds.extend(layer.getBounds());
+                }
+            });
+            if (bounds.isValid()) {
+                {{ this._parent.get_name() }}.{{ this.method }}(bounds, {{ this.options|tojson }});
+            }
+        }
+        {{ this._parent.get_name() }}.on('overlayadd', customFlyToBounds);
+        {%- if this.fit_on_map_load %}
+        customFlyToBounds();
+        {%- endif %}
+        {% endmacro %}
+    """
+    )
+
+    def __init__(
+        self,
+        padding: int = 0,
+        max_zoom: Optional[int] = None,
+        fly: bool = False,
+        fit_on_map_load: bool = True,
+    ):
+        super().__init__()
+        self._name = "FitOverlays"
+        self.method = "flyToBounds" if fly else "fitBounds"
+        self.fit_on_map_load = fit_on_map_load
+        self.options = parse_options(padding=(padding, padding), max_zoom=max_zoom)
+
+
 class CustomPane(MacroElement):
     """
     Creates a custom pane to hold map elements.
@@ -615,7 +722,12 @@ class CustomPane(MacroElement):
         """
     )
 
-    def __init__(self, name, z_index=625, pointer_events=False):
+    def __init__(
+        self,
+        name: str,
+        z_index: Union[int, str] = 625,
+        pointer_events: bool = False,
+    ):
         super().__init__()
         self._name = "Pane"
         self.name = name
