@@ -3,13 +3,20 @@ Wraps leaflet TileLayer, WmsTileLayer (TileLayer.WMS), ImageOverlay, and VideoOv
 
 """
 
+from typing import Any, Callable, Optional, Union
+
+import xyzservices
 from branca.element import Element, Figure
-from jinja2 import Environment, PackageLoader, Template
+from jinja2 import Template
 
 from folium.map import Layer
-from folium.utilities import image_to_url, mercator_transform, parse_options
-
-ENV = Environment(loader=PackageLoader("folium", "templates"))
+from folium.utilities import (
+    TypeBounds,
+    TypeJsonValue,
+    image_to_url,
+    mercator_transform,
+    parse_options,
+)
 
 
 class TileLayer(Layer):
@@ -19,26 +26,29 @@ class TileLayer(Layer):
     Parameters
     ----------
     tiles: str or :class:`xyzservices.TileProvider`, default 'OpenStreetMap'
-        Map tileset to use. Can choose from this list of built-in tiles:
+        Map tileset to use. Folium has built-in all tilesets
+        available in the ``xyzservices`` package. For example, you can pass
+        any of the following to the "tiles" keyword:
+
             - "OpenStreetMap"
-            - "Stamen Terrain", "Stamen Toner", "Stamen Watercolor"
-            - "CartoDB positron", "CartoDB dark_matter"
+            - "CartoDB Positron"
+            - "CartoDB Voyager"
 
-        You can pass a custom tileset to Folium by passing a
+        Explore more provider names available in ``xyzservices`` here:
+        https://leaflet-extras.github.io/leaflet-providers/preview/.
+
+        You can also pass a custom tileset by passing a
         :class:`xyzservices.TileProvider` or a Leaflet-style
-        URL to the tiles parameter: ``http://{s}.yourtiles.com/{z}/{x}/{y}.png``.
-
-        You can find a list of free tile providers here:
-        ``http://leaflet-extras.github.io/leaflet-providers/preview/``.
-        Be sure to check their terms and conditions and to provide attribution
-        with the `attr` keyword.
-    min_zoom: int, default 0
-        Minimum allowed zoom level for this tile layer.
-    max_zoom: int, default 18
-        Maximum allowed zoom level for this tile layer.
+        URL to the tiles parameter: ``https://{s}.yourtiles.com/{z}/{x}/{y}.png``.
+    min_zoom: int, optional, default 0
+        Minimum allowed zoom level for this tile layer. Filled by xyzservices by default.
+    max_zoom: int, optional, default 18
+        Maximum allowed zoom level for this tile layer. Filled by xyzservices by default.
     max_native_zoom: int, default None
         The highest zoom level at which the tile server can provide tiles.
-        If provided you can zoom in past this level. Else tiles will turn grey.
+        Filled by xyzservices by default.
+        By setting max_zoom higher than max_native_zoom, you can zoom in
+        past max_native_zoom, tiles will be autoscaled.
     attr: string, default None
         Map tile attribution; only required if passing custom tile URL.
     detect_retina: bool, default False
@@ -52,7 +62,9 @@ class TileLayer(Layer):
     control : bool, default True
         Whether the Layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
+        When adding multiple base layers, use this parameter to select which one
+        should be shown when opening the map, by not showing the others.
     subdomains: list of strings, default ['abc']
         Subdomains of the tile service.
     tms: bool, default False
@@ -71,37 +83,49 @@ class TileLayer(Layer):
             var {{ this.get_name() }} = L.tileLayer(
                 {{ this.tiles|tojson }},
                 {{ this.options|tojson }}
-            ).addTo({{ this._parent.get_name() }});
+            );
         {% endmacro %}
         """
     )
 
     def __init__(
         self,
-        tiles="OpenStreetMap",
-        min_zoom=0,
-        max_zoom=18,
-        max_native_zoom=None,
-        attr=None,
-        detect_retina=False,
-        name=None,
-        overlay=False,
-        control=True,
-        show=True,
-        no_wrap=False,
-        subdomains="abc",
-        tms=False,
-        opacity=1,
-        **kwargs
+        tiles: Union[str, xyzservices.TileProvider] = "OpenStreetMap",
+        min_zoom: Optional[int] = None,
+        max_zoom: Optional[int] = None,
+        max_native_zoom: Optional[int] = None,
+        attr: Optional[str] = None,
+        detect_retina: bool = False,
+        name: Optional[str] = None,
+        overlay: bool = False,
+        control: bool = True,
+        show: bool = True,
+        no_wrap: bool = False,
+        subdomains: str = "abc",
+        tms: bool = False,
+        opacity: float = 1,
+        **kwargs,
     ):
+        if isinstance(tiles, str):
+            if tiles.lower() == "openstreetmap":
+                tiles = "OpenStreetMap Mapnik"
+                if name is None:
+                    name = "openstreetmap"
+            try:
+                tiles = xyzservices.providers.query_name(tiles)
+            except ValueError:
+                # no match, likely a custom URL
+                pass
 
-        # check for xyzservices.TileProvider without importing it
-        if isinstance(tiles, dict):
-            attr = attr if attr else tiles.html_attribution
-            min_zoom = tiles.get("min_zoom", min_zoom)
-            max_zoom = tiles.get("max_zoom", max_zoom)
+        if isinstance(tiles, xyzservices.TileProvider):
+            attr = attr if attr else tiles.html_attribution  # type: ignore
+            min_zoom = min_zoom or tiles.get("min_zoom")
+            max_zoom = max_zoom or tiles.get("max_zoom")
+            max_native_zoom = max_native_zoom or tiles.get("max_zoom")
             subdomains = tiles.get("subdomains", subdomains)
-            tiles = tiles.build_url(fill_subdomain=False, scale_factor="{r}")
+            if name is None:
+                name = tiles.name.replace(".", "").lower()
+            tiles = tiles.build_url(fill_subdomain=False, scale_factor="{r}")  # type: ignore
 
         self.tile_name = (
             name if name is not None else "".join(tiles.lower().strip().split())
@@ -110,41 +134,22 @@ class TileLayer(Layer):
             name=self.tile_name, overlay=overlay, control=control, show=show
         )
         self._name = "TileLayer"
-        self._env = ENV
 
-        tiles_flat = "".join(tiles.lower().strip().split())
-        if tiles_flat in {"cloudmade", "mapbox", "mapboxbright", "mapboxcontrolroom"}:
-            # added in May 2020 after v0.11.0, remove in a future release
-            raise ValueError(
-                "Built-in templates for Mapbox and Cloudmade have been removed. "
-                "You can still use these providers by passing a URL to the `tiles` "
-                "argument. See the documentation of the `TileLayer` class."
-            )
-        templates = list(
-            self._env.list_templates(filter_func=lambda x: x.startswith("tiles/"))
-        )
-        tile_template = "tiles/" + tiles_flat + "/tiles.txt"
-        attr_template = "tiles/" + tiles_flat + "/attr.txt"
-
-        if tile_template in templates and attr_template in templates:
-            self.tiles = self._env.get_template(tile_template).render()
-            attr = self._env.get_template(attr_template).render()
-        else:
-            self.tiles = tiles
-            if not attr:
-                raise ValueError("Custom tiles must have an attribution.")
+        self.tiles = tiles
+        if not attr:
+            raise ValueError("Custom tiles must have an attribution.")
 
         self.options = parse_options(
-            min_zoom=min_zoom,
-            max_zoom=max_zoom,
-            max_native_zoom=max_native_zoom or max_zoom,
+            min_zoom=min_zoom or 0,
+            max_zoom=max_zoom or 18,
+            max_native_zoom=max_native_zoom,
             no_wrap=no_wrap,
             attribution=attr,
             subdomains=subdomains,
             detect_retina=detect_retina,
             tms=tms,
             opacity=opacity,
-            **kwargs
+            **kwargs,
         )
 
 
@@ -176,7 +181,7 @@ class WmsTileLayer(Layer):
     control : bool, default True
         Whether the Layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
     **kwargs : additional keyword arguments
         Passed through to the underlying tileLayer.wms object and can be used
         for setting extra tileLayer.wms parameters or as extra parameters in
@@ -191,25 +196,25 @@ class WmsTileLayer(Layer):
             var {{ this.get_name() }} = L.tileLayer.wms(
                 {{ this.url|tojson }},
                 {{ this.options|tojson }}
-            ).addTo({{ this._parent.get_name() }});
+            );
         {% endmacro %}
         """
     )  # noqa
 
     def __init__(
         self,
-        url,
-        layers,
-        styles="",
-        fmt="image/jpeg",
-        transparent=False,
-        version="1.1.1",
-        attr="",
-        name=None,
-        overlay=True,
-        control=True,
-        show=True,
-        **kwargs
+        url: str,
+        layers: str,
+        styles: str = "",
+        fmt: str = "image/jpeg",
+        transparent: bool = False,
+        version: str = "1.1.1",
+        attr: str = "",
+        name: Optional[str] = None,
+        overlay: bool = True,
+        control: bool = True,
+        show: bool = True,
+        **kwargs,
     ):
         super().__init__(name=name, overlay=overlay, control=control, show=show)
         self.url = url
@@ -221,7 +226,7 @@ class WmsTileLayer(Layer):
             transparent=transparent,
             version=version,
             attribution=attr,
-            **kwargs
+            **kwargs,
         )
         if cql_filter:
             # special parameter that shouldn't be camelized
@@ -268,7 +273,7 @@ class ImageOverlay(Layer):
     control : bool, default True
         Whether the Layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
 
     See https://leafletjs.com/reference.html#imageoverlay for more
     options.
@@ -282,24 +287,24 @@ class ImageOverlay(Layer):
                 {{ this.url|tojson }},
                 {{ this.bounds|tojson }},
                 {{ this.options|tojson }}
-            ).addTo({{ this._parent.get_name() }});
+            );
         {% endmacro %}
         """
     )
 
     def __init__(
         self,
-        image,
-        bounds,
-        origin="upper",
-        colormap=None,
-        mercator_project=False,
-        pixelated=True,
-        name=None,
-        overlay=True,
-        control=True,
-        show=True,
-        **kwargs
+        image: Any,
+        bounds: TypeBounds,
+        origin: str = "upper",
+        colormap: Optional[Callable] = None,
+        mercator_project: bool = False,
+        pixelated: bool = True,
+        name: Optional[str] = None,
+        overlay: bool = True,
+        control: bool = True,
+        show: bool = True,
+        **kwargs,
     ):
         super().__init__(name=name, overlay=overlay, control=control, show=show)
         self._name = "ImageOverlay"
@@ -308,12 +313,12 @@ class ImageOverlay(Layer):
         self.pixelated = pixelated
         if mercator_project:
             image = mercator_transform(
-                image, [bounds[0][0], bounds[1][0]], origin=origin
+                image, (bounds[0][0], bounds[1][0]), origin=origin
             )
 
         self.url = image_to_url(image, origin=origin, colormap=colormap)
 
-    def render(self, **kwargs):
+    def render(self, **kwargs) -> None:
         super().render()
 
         figure = self.get_root()
@@ -338,7 +343,7 @@ class ImageOverlay(Layer):
                 Element(pixelated), name="leaflet-image-layer"
             )  # noqa
 
-    def _get_self_bounds(self):
+    def _get_self_bounds(self) -> TypeBounds:
         """
         Computes the bounds of the object itself (not including it's children)
         in the form [[lat_min, lon_min], [lat_max, lon_max]].
@@ -367,7 +372,7 @@ class VideoOverlay(Layer):
     control : bool, default True
         Whether the Layer will be included in LayerControls.
     show: bool, default True
-        Whether the layer will be shown on opening (only for overlays).
+        Whether the layer will be shown on opening.
     **kwargs:
         Other valid (possibly inherited) options. See:
         https://leafletjs.com/reference.html#videooverlay
@@ -381,22 +386,22 @@ class VideoOverlay(Layer):
                 {{ this.video_url|tojson }},
                 {{ this.bounds|tojson }},
                 {{ this.options|tojson }}
-            ).addTo({{ this._parent.get_name() }});
+            );
         {% endmacro %}
         """
     )
 
     def __init__(
         self,
-        video_url,
-        bounds,
-        autoplay=True,
-        loop=True,
-        name=None,
-        overlay=True,
-        control=True,
-        show=True,
-        **kwargs
+        video_url: str,
+        bounds: TypeBounds,
+        autoplay: bool = True,
+        loop: bool = True,
+        name: Optional[str] = None,
+        overlay: bool = True,
+        control: bool = True,
+        show: bool = True,
+        **kwargs: TypeJsonValue,
     ):
         super().__init__(name=name, overlay=overlay, control=control, show=show)
         self._name = "VideoOverlay"
@@ -405,7 +410,7 @@ class VideoOverlay(Layer):
         self.bounds = bounds
         self.options = parse_options(autoplay=autoplay, loop=loop, **kwargs)
 
-    def _get_self_bounds(self):
+    def _get_self_bounds(self) -> TypeBounds:
         """
         Computes the bounds of the object itself (not including it's children)
         in the form [[lat_min, lon_min], [lat_max, lon_max]]
