@@ -5,14 +5,35 @@ import json
 import math
 import os
 import re
-import struct
 import tempfile
 import uuid
-import zlib
 from contextlib import contextmanager
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    Union,
+)
 from urllib.parse import urlparse, uses_netloc, uses_params, uses_relative
 
 import numpy as np
+from branca.element import Element, Figure
+
+# import here for backwards compatibility
+from branca.utilities import (  # noqa F401
+    _locations_mirror,
+    _parse_size,
+    none_max,
+    none_min,
+    write_png,
+)
 
 try:
     import pandas as pd
@@ -20,12 +41,23 @@ except ImportError:
     pd = None
 
 
+TypeLine = Iterable[Sequence[float]]
+TypeMultiLine = Union[TypeLine, Iterable[TypeLine]]
+
+TypeJsonValueNoNone = Union[str, float, bool, Sequence, dict]
+TypeJsonValue = Union[TypeJsonValueNoNone, None]
+
+TypePathOptions = Union[bool, str, float, None]
+
+TypeBounds = Sequence[Sequence[float]]
+
+
 _VALID_URLS = set(uses_relative + uses_netloc + uses_params)
 _VALID_URLS.discard("")
 _VALID_URLS.add("data")
 
 
-def validate_location(location):  # noqa: C901
+def validate_location(location: Sequence[float]) -> List[float]:
     """Validate a single lat/lon coordinate pair and convert to a list
 
     Validate that location:
@@ -34,11 +66,6 @@ def validate_location(location):  # noqa: C901
     * allows indexing (i.e. has an ordering)
     * where both values are floats (or convertible to float)
     * and both values are not NaN
-
-    Returns
-    -------
-    list[float, float]
-
     """
     if isinstance(location, np.ndarray) or (
         pd is not None and isinstance(location, pd.DataFrame)
@@ -48,21 +75,19 @@ def validate_location(location):  # noqa: C901
         raise TypeError(
             "Location should be a sized variable, "
             "for example a list or a tuple, instead got "
-            "{!r} of type {}.".format(location, type(location))
+            f"{location!r} of type {type(location)}."
         )
     if len(location) != 2:
         raise ValueError(
             "Expected two (lat, lon) values for location, "
-            "instead got: {!r}.".format(location)
+            f"instead got: {location!r}."
         )
     try:
         coords = (location[0], location[1])
     except (TypeError, KeyError):
         raise TypeError(
             "Location should support indexing, like a list or "
-            "a tuple does, instead got {!r} of type {}.".format(
-                location, type(location)
-            )
+            f"a tuple does, instead got {location!r} of type {type(location)}."
         )
     for coord in coords:
         try:
@@ -70,46 +95,52 @@ def validate_location(location):  # noqa: C901
         except (TypeError, ValueError):
             raise ValueError(
                 "Location should consist of two numerical values, "
-                "but {!r} of type {} is not convertible to float.".format(
-                    coord, type(coord)
-                )
+                f"but {coord!r} of type {type(coord)} is not convertible to float."
             )
         if math.isnan(float(coord)):
             raise ValueError("Location values cannot contain NaNs.")
     return [float(x) for x in coords]
 
 
-def validate_locations(locations):
-    """Validate an iterable with multiple lat/lon coordinate pairs.
-
-    Returns
-    -------
-    list[list[float, float]] or list[list[list[float, float]]]
-
-    """
-    locations = if_pandas_df_convert_to_numpy(locations)
+def _validate_locations_basics(locations: TypeMultiLine) -> None:
+    """Helper function that does basic validation of line and multi-line types."""
     try:
         iter(locations)
     except TypeError:
         raise TypeError(
             "Locations should be an iterable with coordinate pairs,"
-            " but instead got {!r}.".format(locations)
+            f" but instead got {locations!r}."
         )
     try:
         next(iter(locations))
     except StopIteration:
         raise ValueError("Locations is empty.")
+
+
+def validate_locations(locations: TypeLine) -> List[List[float]]:
+    """Validate an iterable with lat/lon coordinate pairs."""
+    locations = if_pandas_df_convert_to_numpy(locations)
+    _validate_locations_basics(locations)
+    return [validate_location(coord_pair) for coord_pair in locations]
+
+
+def validate_multi_locations(
+    locations: TypeMultiLine,
+) -> Union[List[List[float]], List[List[List[float]]]]:
+    """Validate an iterable with possibly nested lists of coordinate pairs."""
+    locations = if_pandas_df_convert_to_numpy(locations)
+    _validate_locations_basics(locations)
     try:
-        float(next(iter(next(iter(next(iter(locations)))))))
+        float(next(iter(next(iter(next(iter(locations)))))))  # type: ignore
     except (TypeError, StopIteration):
         # locations is a list of coordinate pairs
-        return [validate_location(coord_pair) for coord_pair in locations]
+        return [validate_location(coord_pair) for coord_pair in locations]  # type: ignore
     else:
         # locations is a list of a list of coordinate pairs, recurse
-        return [validate_locations(lst) for lst in locations]
+        return [validate_locations(lst) for lst in locations]  # type: ignore
 
 
-def if_pandas_df_convert_to_numpy(obj):
+def if_pandas_df_convert_to_numpy(obj: Any) -> Any:
     """Return a Numpy array from a Pandas dataframe.
 
     Iterating over a DataFrame has weird side effects, such as the first
@@ -121,7 +152,11 @@ def if_pandas_df_convert_to_numpy(obj):
         return obj
 
 
-def image_to_url(image, colormap=None, origin="upper"):
+def image_to_url(
+    image: Any,
+    colormap: Optional[Callable] = None,
+    origin: str = "upper",
+) -> str:
     """
     Infers the type of an image argument and transforms it into a URL.
 
@@ -159,7 +194,7 @@ def image_to_url(image, colormap=None, origin="upper"):
     return url.replace("\n", " ")
 
 
-def _is_url(url):
+def _is_url(url: str) -> bool:
     """Check to see if `url` has a valid protocol."""
     try:
         return urlparse(url).scheme in _VALID_URLS
@@ -167,99 +202,12 @@ def _is_url(url):
         return False
 
 
-def write_png(data, origin="upper", colormap=None):
-    """
-    Transform an array of data into a PNG string.
-    This can be written to disk using binary I/O, or encoded using base64
-    for an inline PNG like this:
-
-    >>> png_str = write_png(array)
-    >>> "data:image/png;base64," + png_str.encode("base64")
-
-    Inspired from
-    https://stackoverflow.com/questions/902761/saving-a-numpy-array-as-an-image
-
-    Parameters
-    ----------
-    data: numpy array or equivalent list-like object.
-         Must be NxM (mono), NxMx3 (RGB) or NxMx4 (RGBA)
-
-    origin : ['upper' | 'lower'], optional, default 'upper'
-        Place the [0,0] index of the array in the upper left or lower left
-        corner of the axes.
-
-    colormap : callable, used only for `mono` image.
-        Function of the form [x -> (r,g,b)] or [x -> (r,g,b,a)]
-        for transforming a mono image into RGB.
-        It must output iterables of length 3 or 4, with values between
-        0. and 1.  Hint: you can use colormaps from `matplotlib.cm`.
-
-    Returns
-    -------
-    PNG formatted byte string
-
-    """
-    if colormap is None:
-
-        def colormap(x):
-            return (x, x, x, 1)
-
-    arr = np.atleast_3d(data)
-    height, width, nblayers = arr.shape
-
-    if nblayers not in [1, 3, 4]:
-        raise ValueError("Data must be NxM (mono), NxMx3 (RGB), or NxMx4 (RGBA)")
-    assert arr.shape == (height, width, nblayers)
-
-    if nblayers == 1:
-        arr = np.array(list(map(colormap, arr.ravel())))
-        nblayers = arr.shape[1]
-        if nblayers not in [3, 4]:
-            raise ValueError(
-                "colormap must provide colors of length 3 (RGB) or 4 (RGBA)"
-            )
-        arr = arr.reshape((height, width, nblayers))
-    assert arr.shape == (height, width, nblayers)
-
-    if nblayers == 3:
-        arr = np.concatenate((arr, np.ones((height, width, 1))), axis=2)
-        nblayers = 4
-    assert arr.shape == (height, width, nblayers)
-    assert nblayers == 4
-
-    # Normalize to uint8 if it isn't already.
-    if arr.dtype != "uint8":
-        with np.errstate(divide="ignore", invalid="ignore"):
-            arr = arr * 255.0 / arr.max(axis=(0, 1)).reshape((1, 1, 4))
-            arr[~np.isfinite(arr)] = 0
-        arr = arr.astype("uint8")
-
-    # Eventually flip the image.
-    if origin == "lower":
-        arr = arr[::-1, :, :]
-
-    # Transform the array to bytes.
-    raw_data = b"".join([b"\x00" + arr[i, :, :].tobytes() for i in range(height)])
-
-    def png_pack(png_tag, data):
-        chunk_head = png_tag + data
-        return (
-            struct.pack("!I", len(data))
-            + chunk_head
-            + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
-        )
-
-    return b"".join(
-        [
-            b"\x89PNG\r\n\x1a\n",
-            png_pack(b"IHDR", struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
-            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
-            png_pack(b"IEND", b""),
-        ]
-    )
-
-
-def mercator_transform(data, lat_bounds, origin="upper", height_out=None):
+def mercator_transform(
+    data: Any,
+    lat_bounds: Tuple[float, float],
+    origin: str = "upper",
+    height_out: Optional[int] = None,
+) -> np.ndarray:
     """
     Transforms an image computed in (longitude,latitude) coordinates into
     the a Mercator projection image.
@@ -286,7 +234,6 @@ def mercator_transform(data, lat_bounds, origin="upper", height_out=None):
     See https://en.wikipedia.org/wiki/Web_Mercator for more details.
 
     """
-    import numpy as np
 
     def mercator(x):
         return np.arcsinh(np.tan(x * np.pi / 180.0)) * 180.0 / np.pi
@@ -321,25 +268,7 @@ def mercator_transform(data, lat_bounds, origin="upper", height_out=None):
     return out
 
 
-def none_min(x, y):
-    if x is None:
-        return y
-    elif y is None:
-        return x
-    else:
-        return min(x, y)
-
-
-def none_max(x, y):
-    if x is None:
-        return y
-    elif y is None:
-        return x
-    else:
-        return max(x, y)
-
-
-def iter_coords(obj):
+def iter_coords(obj: Any) -> Iterator[Tuple[float, ...]]:
     """
     Returns all the coordinate tuples from a geometry or feature.
 
@@ -347,10 +276,18 @@ def iter_coords(obj):
     if isinstance(obj, (tuple, list)):
         coords = obj
     elif "features" in obj:
-        coords = [geom["geometry"]["coordinates"] for geom in obj["features"]]
+        coords = [
+            geom["geometry"]["coordinates"]
+            for geom in obj["features"]
+            if geom["geometry"]
+        ]
     elif "geometry" in obj:
-        coords = obj["geometry"]["coordinates"]
-    elif "geometries" in obj and "coordinates" in obj["geometries"][0]:
+        coords = obj["geometry"]["coordinates"] if obj["geometry"] else []
+    elif (
+        "geometries" in obj
+        and obj["geometries"][0]
+        and "coordinates" in obj["geometries"][0]
+    ):
         coords = obj["geometries"][0]["coordinates"]
     else:
         coords = obj.get("coordinates", obj)
@@ -362,30 +299,16 @@ def iter_coords(obj):
             yield from iter_coords(coord)
 
 
-def _locations_mirror(x):
-    """
-    Mirrors the points in a list-of-list-of-...-of-list-of-points.
-    For example:
-    >>> _locations_mirror([[[1, 2], [3, 4]], [5, 6], [7, 8]])
-    [[[2, 1], [4, 3]], [6, 5], [8, 7]]
-
-    """
-    if hasattr(x, "__iter__"):
-        if hasattr(x[0], "__iter__"):
-            return list(map(_locations_mirror, x))
-        else:
-            return list(x[::-1])
-    else:
-        return x
-
-
-def get_bounds(locations, lonlat=False):
+def get_bounds(
+    locations: Any,
+    lonlat: bool = False,
+) -> List[List[Optional[float]]]:
     """
     Computes the bounds of the object in the form
     [[lat_min, lon_min], [lat_max, lon_max]]
 
     """
-    bounds = [[None, None], [None, None]]
+    bounds: List[List[Optional[float]]] = [[None, None], [None, None]]
     for point in iter_coords(locations):
         bounds = [
             [
@@ -402,7 +325,7 @@ def get_bounds(locations, lonlat=False):
     return bounds
 
 
-def camelize(key):
+def camelize(key: str) -> str:
     """Convert a python_style_variable_name to lowerCamelCase.
 
     Examples
@@ -415,23 +338,7 @@ def camelize(key):
     return "".join(x.capitalize() if i > 0 else x for i, x in enumerate(key.split("_")))
 
 
-def _parse_size(value):
-    try:
-        if isinstance(value, (int, float)):
-            value_type = "px"
-            value = float(value)
-            assert value > 0
-        else:
-            value_type = "%"
-            value = float(value.strip("%"))
-            assert 0 <= value <= 100
-    except Exception:
-        msg = "Cannot parse value {!r} as {!r}".format
-        raise ValueError(msg(value, value_type))
-    return value, value_type
-
-
-def compare_rendered(obj1, obj2):
+def compare_rendered(obj1: str, obj2: str) -> bool:
     """
     Return True/False if the normalized rendered version of
     two folium map objects are the equal or not.
@@ -440,7 +347,7 @@ def compare_rendered(obj1, obj2):
     return normalize(obj1) == normalize(obj2)
 
 
-def normalize(rendered):
+def normalize(rendered: str) -> str:
     """Return the input string without non-functional spaces or newlines."""
     out = "".join([line.strip() for line in rendered.splitlines() if line.strip()])
     out = out.replace(", ", ",")
@@ -448,7 +355,7 @@ def normalize(rendered):
 
 
 @contextmanager
-def temp_html_filepath(data):
+def temp_html_filepath(data: str) -> Iterator[str]:
     """Yields the path of a temporary HTML file containing data."""
     filepath = ""
     try:
@@ -461,7 +368,7 @@ def temp_html_filepath(data):
             os.remove(filepath)
 
 
-def deep_copy(item_original):
+def deep_copy(item_original: Element) -> Element:
     """Return a recursive deep-copy of item where each copy has a new ID."""
     item = copy.copy(item_original)
     item._id = uuid.uuid4().hex
@@ -475,30 +382,62 @@ def deep_copy(item_original):
     return item
 
 
-def get_obj_in_upper_tree(element, cls):
+def get_obj_in_upper_tree(element: Element, cls: Type) -> Element:
     """Return the first object in the parent tree of class `cls`."""
-    if not hasattr(element, "_parent"):
-        raise ValueError(f"The top of the tree was reached without finding a {cls}")
     parent = element._parent
+    if parent is None:
+        raise ValueError(f"The top of the tree was reached without finding a {cls}")
     if not isinstance(parent, cls):
         return get_obj_in_upper_tree(parent, cls)
     return parent
 
 
-def parse_options(**kwargs):
+def parse_options(**kwargs: TypeJsonValue) -> Dict[str, TypeJsonValueNoNone]:
     """Return a dict with lower-camelcase keys and non-None values.."""
     return {camelize(key): value for key, value in kwargs.items() if value is not None}
 
 
-def escape_backticks(text):
+def escape_backticks(text: str) -> str:
     """Escape backticks so text can be used in a JS template."""
     return re.sub(r"(?<!\\)`", r"\`", text)
 
 
-def escape_double_quotes(text):
+def escape_double_quotes(text: str) -> str:
     return text.replace('"', r"\"")
 
 
-def javascript_identifier_path_to_array_notation(path):
+def javascript_identifier_path_to_array_notation(path: str) -> str:
     """Convert a path like obj1.obj2 to array notation: ["obj1"]["obj2"]."""
     return "".join(f'["{escape_double_quotes(x)}"]' for x in path.split("."))
+
+
+def get_and_assert_figure_root(obj: Element) -> Figure:
+    """Return the root element of the tree and assert it's a Figure."""
+    figure = obj.get_root()
+    assert isinstance(
+        figure, Figure
+    ), "You cannot render this Element if it is not in a Figure."
+    return figure
+
+
+class JsCode:
+    """Wrapper around Javascript code."""
+
+    def __init__(self, js_code: Union[str, "JsCode"]):
+        if isinstance(js_code, JsCode):
+            self.js_code: str = js_code.js_code
+        else:
+            self.js_code = js_code
+
+    def __str__(self):
+        return self.js_code
+
+
+def parse_font_size(value: Union[str, int, float]) -> str:
+    """Parse a font size value, if number set as px"""
+    if isinstance(value, (int, float)):
+        return f"{value}px"
+
+    if (value[-3:] != "rem") and (value[-2:] not in ["em", "px"]):
+        raise ValueError("The font size must be expressed in rem, em, or px.")
+    return value
